/**
 * option keys:
 *   items:        The items that will be displayed. This must be an 
 *                   array. The array can contain either strings or objects.
 *                   See below for the item object structure.
 *   center:       An array containing the x and y center for this menu.
 *                   Defaults to [Renderer.screen.getWidth() / 2, 
 *                   Renderer.screen.getHeight() / 2]. The array can also 
 *                   contain functions for dynamic centering.
 *   innerRadius:  A number or function defining the inner circle's radius. 
 *                   Defaults to Renderer.screen.getHeight() / 8.
 *   outerRadius:  A number or function defining the outer circle's radius. 
 *                   Defaults to Renderer.screen.getHeight() / 2.7.
 *   rotOffset:    A rotational offset, in radians. Defaults to 0.
 *   clockwise:    If true, draws the strings clockwise.
 *   bgColor:      Option background color. Defaults to
 *                   Renderer.color(50, 50, 50, 150).
 *   fgColor:      Option foreground color (the color drawn when an option
 *                   is hovered). Defaults to 
 *                   Renderer.color(150, 150, 150, 255).
 *   lineSepColor: Option seperator color. Defaults to 
 *                   Renderer.color(50, 50, 50, 125). Drawn over option
 *                   backgrounds.
 *   callback:     Callback function if an option is pressed. Passed the
 *                   clicked item as the first argument, and the selector
 *                   as the second argument.
 * 
 * Providing an item as an object rather than a string allows more 
 * fine-grained control over the display of that particular item. No matter
 * the type of the item, that exact item will be returned to the callback
 * function if it is clicked. This means that you can supply arbitrary
 * data in the item object, and receive that data in the callback.
 * 
 * If the item is given as an object, you can also include the following
 * keys:
 *   width:        Specify a custom width, in radians. Must be less 
 *                   than 2 * PI.
 *   key:          A Keyboard key. If provided, when this key is pressed,
 *                   it's item will be selected. An alternative to using
 *                   the mouse to click the option.
 *   text:         Contains the name to display. Can be a string or a Text
 *                   object.
 *   innerRadius:  See above.
 *   outerRadius:  See above.
 *   bgColor:      See above.
 *   fgColor:      See above.
 *   callback:     Takes precedence over the general callback provided
 *                 to the RingSelector. If an item callback is supplied,
 *                 the general callback will NOT be called for that item.
 */
function RingSelector(options) {
  this.TWO_PI = 2 * Math.PI;

  this.items = options.items;
  this.rotOffset = options.rotOffset || 0;

  this.hoveredOption = undefined;
  this.opened = false;

  this.xCenter = function() {
    var c = (options.center && options.center[0]) || Renderer.screen.getWidth() / 2;
    return typeof c === 'function' ? c() : c;
  }

  this.yCenter = function() {
    var c = (options.center && options.center[1]) || Renderer.screen.getHeight() / 2;
    return typeof c === 'function' ? c() : c;
  }

  this.mappedItems = this.items.map(function(item) {
    if (!(typeof item === 'string' || typeof item === 'object')) 
      throw new Error('RingSelector: item must be a string or an object');

    var name = item.name || item;
    if (!(name instanceof Text)) 
      name = new Text(name, 0, 0);

    function innerRadius() {
      var r = Renderer.screen.getHeight() / 8;

      if (options.innerRadius !== undefined) r = options.innerRadius;
      if (item.innerRadius !== undefined) r = item.innerRadius;

      return typeof r === 'function' ? r() : r;
    }

    function outerRadius() {
      var r = Renderer.screen.getHeight() / 2.7;

      if (options.outerRadius !== undefined) r = options.outerRadius;
      if (item.outerRadius !== undefined) r = item.outerRadius;

      return typeof r === 'function' ? r() : r;
    }

    return {
      name: name,
      innerRadius: innerRadius,
      outerRadius: outerRadius,
      bgColor: item.bgColor || options.bgColor || Renderer.color(50, 50, 50, 150),
      fgColor: item.fgColor || options.fgColor || Renderer.color(175, 175, 175, 200),
      color: item.bgColor   || options.bgColor || Renderer.color(50, 50, 50, 150),
      width: item.width,
      key: item.key,
      callback: item.callback
    };
  });

  // Handle the item 'width' key
  (function() {
    var totalWidth = this.TWO_PI;
    var itemsLeft = this.mappedItems.length;

    this.mappedItems.forEach(function(item) {
      var width = item.width;
      
      if (width) {
        if (width > this.TWO_PI) 
          throw new Error("RingSelector: item 'width' key must be less than 2*PI");
          
        totalWidth -= width;
        itemsLeft--;
        
        if (totalWidth >= this.TWO_PI && itemsLeft > 0) 
        throw new Error("RingSelector: radius required to display all items exceeds 2*PI");
      }
    });

    var prev = 0;

    this.mappedItems.forEach(function(item, i) {
      var width = item.width || (totalWidth / itemsLeft);
  
      this.mappedItems[i].startAngle = prev;
      this.mappedItems[i].endAngle = prev + width;
      prev += width;
    }.bind(this));
  }.bind(this))();

  // Reverse if clockwise
  if (options.clockwise) {
    this.items = this.items.reverse();
    this.mappedItems = this.mappedItems.reverse();
  }
  
  this.open = function() {
    this.gui.open();
  }.bind(this);

  this.close = function() {
    this.gui.close();
  }.bind(this);

  this.normalizeAngle = function(theta) {
    while (theta < 0) 
      theta += this.TWO_PI;
    
    while (theta >= this.TWO_PI) 
      theta -= this.TWO_PI;
  
    return theta;
  }

  this.isOptionHovered = function(x, y, item) {
    x -= this.xCenter();
    y -= this.yCenter();

    var hypot = Math.sqrt(Math.pow(x, 2) + Math.pow(y, 2));
    var mAngle = this.normalizeAngle(Math.atan2(y, x) + this.rotOffset);

    return mAngle >= item.startAngle && mAngle < item.endAngle &&
           hypot >= item.innerRadius() && hypot <= item.outerRadius();
  }.bind(this);

  this.updateOptionColors = function() {
    if (!this.gui.isOpen() && !this.opened) return;

    this.mappedItems.forEach(function(item, i) {
      this.mappedItems[i].color = easeColor(
        item.color,
        i === this.hoveredOption ? item.fgColor : item.bgColor,
        10, 1
      );
    }.bind(this));
  }.bind(this);

  this.drawAnnularSector = function(options) {
    var item = options.item;

    var endAngle = options.startAngle;
    var startAngle = options.endAngle;
    var innerRadius = item.innerRadius();
    var outerRadius = item.outerRadius();
    var steps = 50;
    var color = item.color || Renderer.color(255, 255, 255, 255);
    var xCenter = this.xCenter();
    var yCenter = this.yCenter();

    var theta = -(endAngle - startAngle) / steps;

    for (var i = 0; i < steps; i++) {
      var p1 = {
        x: xCenter + outerRadius * Math.cos(startAngle - theta * i),
        y: yCenter + outerRadius * Math.sin(startAngle - theta * i)
      };
      var p2 = {
        x: xCenter + outerRadius * Math.cos(startAngle - theta * (i + 1)),
        y: yCenter + outerRadius * Math.sin(startAngle - theta * (i + 1))
      };
      var p3 = {
        x: xCenter + innerRadius * Math.cos(startAngle - theta * (i + 1)),
        y: yCenter + innerRadius * Math.sin(startAngle - theta * (i + 1))
      };
      var p4 = {
        x: xCenter + innerRadius * Math.cos(startAngle - theta * i),
        y: yCenter + innerRadius * Math.sin(startAngle - theta * i)
      };

      new Shape(color)
        .setDrawMode(7)
        .addVertex(p1.x, p1.y)
        .addVertex(p2.x, p2.y)
        .addVertex(p3.x, p3.y)
        .addVertex(p4.x, p4.y)
        .draw();
    }
  }.bind(this);
  
  this.draw = function(mouseX, mouseY) {
    var hovered = false;

    this.mappedItems.forEach(function(item, i) {
      if (this.isOptionHovered(mouseX, mouseY, item)) {
        this.hoveredOption = i;
        hovered = true;
      }
    }.bind(this));

    if (!hovered) this.hoveredOption = undefined;

    this.mappedItems.forEach(function(item) {
      var startAngle = item.startAngle - this.rotOffset;
      var endAngle = item.endAngle - this.rotOffset;
      var midAngle = ((endAngle - startAngle) / 2) + startAngle;
      var innerRadius = item.innerRadius();
      var outerRadius = item.outerRadius();

      var center = {
        x: this.xCenter() + Math.cos(midAngle) * (innerRadius + outerRadius) / 2,
        y: this.yCenter() + Math.sin(midAngle) * (innerRadius + outerRadius) / 2
      };

      this.drawAnnularSector({
        startAngle: startAngle,
        endAngle: endAngle,
        item: item
      });
      
      item.name
        .setAlign('CENTER')
        .setX(center.x)
        .setY(center.y - 4.5)
        .draw();

      var boundary1 = {
        x: this.xCenter() + innerRadius * Math.cos(startAngle),
        y: this.yCenter() + innerRadius * Math.sin(startAngle)
      };
      
      var boundary2 = {
        x: this.xCenter() + outerRadius * Math.cos(startAngle),
        y: this.yCenter() + outerRadius * Math.sin(startAngle)
      };

      Renderer.drawLine(
        options.lineSepColor || Renderer.color(50, 50, 50, 125),
        boundary1.x, boundary1.y,
        boundary2.x, boundary2.y,
        1
      );
    }.bind(this));
  }.bind(this);

  this.clicked = function() {
    var item = this.items[this.hoveredOption];

    if (item)
      (item.callback || options.callback || function(){})(item, this.gui);
  }.bind(this);

  this.keyTyped = function(_, key) {
    var item = this.mappedItems.find(function(item) {
      return item.key === key;
    });

    if (item)
      (item.callback || options.callback || function(){})(item, this.gui);
  }.bind(this);

  this.gui = new Gui();

  this.gui.registerDraw(this.draw);
  this.gui.registerClicked(this.clicked);
  this.gui.registerKeyTyped(this.keyTyped);

  register('step', this.updateOptionColors).setFps(144);
}